/** 
 *   FIXME ....
 *
 *   Commands definitions for "antenna" position system.
 *
 *   (c) Copyright 2003 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 *   TODO: ??? 
 **/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <malloc.h>

#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/errno.h>

#include "command.h"
#include "encoders.h"
#include "iar.h"
#include "coordinate.h"
#include "my_stdio.h"


/* A structure which contains information on the commands this program
   can understand. */

struct com {
	char *name;		/* User printable name of the function. */
	/* Function to call to do the job. */
	int (*func)(int fd_out, char *name);
	char *doc;		/* Documentation for this function.  */
	struct com *extended;
};

typedef struct com COMMAND;

/* The names of functions that actually do the manipulation. */
int com_help(int fd_out, char *);
int com_quit(int fd_out, char *);
int com_getraw(int fd_out, char *);
int com_toggle_echo(int fd_out, char *);
int com_set(int fd_out, char * name);
int com_set_zero(int fd_out, char * name);
int com_set_cw(int fd_out, char * name);
int com_set_ccw(int fd_out, char * name);

COMMAND commands_set[] = {
	{ "zero", com_set_zero, 
	  "Set zero for current position on encoder \"dec\" or \"ra\"" },
	{ "cw", com_set_cw, 
	  "Set clockwhise increment on encoder \"dec\" or \"ra\"" },
	{ "ccw", com_set_ccw, 
	  "Set counter clockwhise increment on encoder \"dec\" or \"ra\"" },
	{ (char *)NULL, (int) NULL, (char *)NULL }
};


COMMAND commands_main[] = {
	{ "set", com_set, 
	  "Set some encoders attributes (see help)", commands_set},
	{ "getenc", com_getraw, "Get raw position of encoders" },
	{ "e", com_toggle_echo, "Toggle remote echo"},
	{ "help", com_help, "Display this text"},
	{ "?", com_help, "Synonym for `help'"},
	{ "quit", com_quit, "Quit ..."},
	{ "q", com_quit, "Quit ..."},
	{ (char *)NULL, (int) NULL, (char *)NULL}
};


/* When non-zero, this means the user is done using this program. */
int done = 0;
int remote_echo = 0;

/* Look up NAME as the name of a command, and return a pointer to that
   command.  Return a NULL pointer if NAME isn't a command name. */
COMMAND *find_command (COMMAND *cm, char *name)
{
	int i;

	for (i = 0; cm[i].name; i++)
		if (strcmp (name, cm[i].name) == 0)
			return (&cm[i]);
	
	return ((COMMAND *)NULL);
}






/* Execute a command line. */
int execute_line (int fd_out, COMMAND *cm, char *line)
{
	int i;
	COMMAND *command;
	char *word;
	
	/* Isolate the command word. */
	i = 0;
	while (line[i] && whitespace (line[i]))	i++;
	word = line + i;
	
	while (line[i] && !whitespace (line[i])) i++;

	if (line[i])
		line[i++] = '\0';

	command = find_command (cm, word);

	if (!command) {
		fd_printf (fd_out, 
			   "# %s: No such command for ""FIXME"".\n", word);
		return (-1);
	}

	/* Get argument to command, if any. */
	while (whitespace (line[i])) i++;

	word = line + i;

	/* Call the function. */
	return ((*(command->func)) (fd_out, word));
}


/* Strip whitespace from the start and end of STRING.  Return a pointer
   into STRING. */
char * stripwhite (char *string)
{
	char *s, *t;

	for (s = string; whitespace (*s); s++);
    
	if (*s == 0)
		return (s);

	t = s + strlen (s) - 1;
	while (t > s && whitespace (*t))
		t--;
	*++t = '\0';
	
	return s;
}


/* The user wishes to quit using this program.  Just set DONE
   non-zero. */
int com_quit (int fd_out, char *arg)
{
	done = 1;
	return (0);
}


int com_toggle_echo (int fd_out, char *arg)
{
	remote_echo ^= 1;
	return (0);
}


/* Print out help for ARG, or for all of the commands if ARG is not
   present. */

int com_help_real (int fd_out, char *arg, COMMAND *cm)
{
	int i, j;
	int printed = 0;
	int cmp;
	char *sub_arg = "";
	COMMAND *commands = cm;
	
	for (i = 0; commands[i].name; i++) {
		cmp = strcmp (arg, commands[i].name);
		if (!*arg || ( cmp == 0)) {
			fd_printf (fd_out, "# %-13s%s.\n", commands[i].name, 
				commands[i].doc);
			printed++;
		}
		if ( commands[i].extended && (cmp >= 0)){
			j = -1;
			if (cmp > 0){
				j = strlen(commands[i].name) + 1;
				while (whitespace (arg[j])) j++;
				sub_arg = arg + j;
			}
			if ( (j + 1) < strlen(arg) ){
				if ( sub_arg == "" )
					fd_printf (fd_out,
						   "# Possible %s options ar:\n", arg);
				com_help_real(fd_out, sub_arg, commands->extended);
				return 0;
			}
		}
	}

	if (!printed) {
		fd_printf (fd_out, 
			   "# No commands match `%s'. Possibilties are:\n", 
			   arg);

		fd_printf (fd_out, "# ");
		for (i = 0; commands[i].name; i++) {
			/* Print in six columns. */
			if (printed == 6) {
				printed = 0;
				fd_printf (fd_out, "\n# ");
			}
			
			fd_printf (fd_out, "%-13s", commands[i].name);
			printed++;
		}

		if (printed)
			fd_printf (fd_out, "\n");
	}
	return (0);
}


/* help front end */

int com_help (int fd_out, char *arg)
{
	return com_help_real(fd_out, arg, commands_main);
}




int command_main(int fd_in, int fd_out)
{
	char *line, *s, c;
	int l;

	done = 0;
	line = (char *) malloc(COMMAND_LINE_LEGEND*sizeof(char));
	/* Loop reading and executing lines until the user quits. */
	while( !done ) {
		//l = read(fd_in, line, COMMAND_LINE_LEGEND);
		l = 0;
		do {
			/* Win telnet clients sen char by char ... */
			read(fd_in, &c, 1); 
			*(line+l) = c;
			l++;
			if ( remote_echo ) fd_printf(fd_out,"%c", c);
		} while(c != '\n' && l < COMMAND_LINE_LEGEND);
		//*(line+l-1)=0;
		if ( *(line+l-2) == 0xd ) *(line+l-2) = 0;
		else *(line+l-1) = 0;
		if (!line)
			break;
		/* Remove leading and trailing whitespace from the
		   line.  Then, if there is anything left, add it to
		   the history list and execute it. */
		s = stripwhite (line);
		if (*s){
			execute_line (fd_out, commands_main, s);
		}
		
	}
	free (line);
	/* fd_printf (fd_out, "# connection closed\n"); */

	return 0;
}
/******************************************************************************
 * Functions for position system:
 *****************************************************************************/


int com_set(int fd_out, char * name)
{
	if ( *name == '\0') {
		fd_printf (fd_out, "# set: missing argument.\n");
		return (-1);
	}
	return execute_line(fd_out, commands_set, name);
}


int com_getraw(int fd_out, char * name)
{
	fd_printf(fd_out,"%5d %5d\n", 
		  get_encoder(0), get_encoder(1) );

	return 0;
}


int com_set_zero(int fd_out, char * name)
{
	char *cm;

	cm = stripwhite(name);
	switch( str2cord_type(cm) ){
	case DEC_CASE:
		fd_printf(fd_out,"# set declination encoder to 0.\n");
		return encoder_set_zero(encoder_num[DEC_CASE]);
	case RA_CASE:
		fd_printf(fd_out,"# set right accession encoder to 0.\n");
		return encoder_set_zero(encoder_num[RA_CASE]);
	default:
		fd_printf(fd_out,"# set zero: argument error: %s\n", cm);
		return -1;
	}
}


int com_set_cw(int fd_out, char * name)
{
	char *cm;

	cm = stripwhite(name);
	switch( str2cord_type(cm) ){
	case DEC_CASE:
		fd_printf(fd_out,
			  "# set dec. encoder to clockwhise increment.\n");
		return encoder_set_cw(encoder_num[DEC_CASE]);
	case RA_CASE:
		fd_printf(fd_out,
			  "# set ra encoder to clockwhise increment.\n");
		return encoder_set_cw(encoder_num[RA_CASE]);
	default:
		fd_printf(fd_out,"# set cw: argument error: %s\n", cm);
		return -1;
	}
}


int com_set_ccw(int fd_out, char * name)
{
	char *cm;

	cm = stripwhite(name);
	switch( str2cord_type(cm) ){
	case DEC_CASE:
		fd_printf(fd_out,
			  "# set dec. encoder to counter clockwhise increment.\n");
		return encoder_set_ccw(encoder_num[DEC_CASE]);
	case RA_CASE:
		fd_printf(fd_out,
			  "# set ra encoder to counter clockwhise increment.\n");
		return encoder_set_ccw(encoder_num[RA_CASE]);
	default:
		fd_printf(fd_out,"# set ccw: argument error: %s\n", cm);
		return -1;
	}
}
